/******************************************************************************
 *
 *              Main Demonstration of fatfs on PIC32 Starter Kit
 *
 ******************************************************************************
 * FileName:        main.c
 * Dependencies:    See include below
 * Processor:       PIC32
 * Compiler:        C32 V1.03
 * Company:         Microchip Technology, Inc.
 *
 * Software License Agreement
 * Microchip Technology Incorporated ("Microchip) and its licensors retain 
 * all ownership and intellectual property rights in the code accompanying 
 * this message and in all derivatives thereto.  The code and information 
 * accompanying this message is provided for your convenience and use with 
 * Microchip products only.  Your acceptance and/or use of this code 
 * constitute agreement to the terms and conditions of this notice.Microchip 
 * disclaims all liability arising from this information and its use.  It is 
 * your responsibility to ensure that your application meets with your 
 * specifications.Further, the implementation and use of the FAT file system 
 * accompanying this code, SD card specifications, MMC card specifications 
 * and other third party tools may require a license from various entities, 
 * including, but not limited to Microsoft Corporation, SD Card Association 
 * and MMCA.  It is your responsibility to obtain more information regarding 
 * any applicable licensing obligations.
 * 
 * CODE AND INFORMATION ACCOMPANYING THIS MESSAGE IS PROVIDED AS IS.  
 * MICROCHIP AND ITS LICENSORS MAKES NO REPRESENTATION OR WARRANTIES OF ANY 
 * KIND WHETHER EXPRESS OR IMPLIED, WRITTEN OR ORAL, STATUTORY OR OTHERWISE, 
 * RELATED TO THE INFORMATION PROVIDED TO YOU, INCLUDING BUT NOT LIMITED TO 
 * ITS CONDITION, QUALITY, PERFORMANCE, MERCHANTABILITY, NON-INFRINGEMENT, 
 * OR FITNESS FOR PURPOSE.  MICROCHIP AND ITS LICENSORS ARE NOT LIABLE, UNDER 
 * ANY CIRCUMSTANCES, FOR SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, 
 * FOR ANY REASON WHATSOEVER.  	 
 *
 *****************************************************************************
 *
 * 11/17/08	V1.00 D.Wenn		First version, designed for PIC32 Starter Kit
 * 11/25/08 V1.01 D.Wenn		Added PIC32 image output
 *
 *****************************************************************************/

#include <plib.h>
#include "HardwareProfile.h"
#include "db_utils.h"
#include <FatFs\diskio.h>
#include <FatFs\ff.h>

//configure the hardware
#pragma config FPLLMUL = MUL_20, FPLLIDIV = DIV_2, FPLLODIV = DIV_1, FWDTEN = OFF
#pragma config POSCMOD = HS, FNOSC = PRIPLL, FPBDIV = DIV_1


///////////////////////////////////////////////////////////////////
// local variables
volatile BYTE rtcYear = 108, rtcMon = 11, rtcMday = 17;	// RTC starting values
volatile BYTE rtcHour = 12, rtcMin, rtcSec;
volatile BYTE tick;			// crude state machine pace control
DWORD acc_size;				/* Work register for fs command */
WORD acc_files, acc_dirs;
FILINFO finfo;

FATFS fatfs[_DRIVES];	// file system object
BYTE buff[1024];		// working buffer
volatile UINT Timer;		/* 1kHz increment timer */

typedef enum {
	WAIT_FOR_DISK,		// wait for disk/card insertion
	IDLE,				// waiting for buttons to be pressed
	ACTION,				// action menu (output bmp, flash leds)
	LED_PLAYBACK_OPEN,	// play led sequence
	LED_PLAYBACK_READ,
	LED_PLAYBACK_CLOSE,
} DEMO_STATE;

// the bitmap image
extern const unsigned char bmpPIC32[230454];

///////////////////////////////////////////////////////////////////
// local functions
static FRESULT scan_files (char* path);

///////////////////////////////////////////////////////////////////

int main(void)
{
	DWORD param1, s1, s2;
	FATFS *fs;
	FIL hFile;
	DIR dir;
	DEMO_STATE demoState = WAIT_FOR_DISK;
	UINT cnt;
	BYTE led_pattern, res;
	unsigned long fpb;
	
	// set the processor for maximum performance
	SYSTEMConfigPerformance(FOSC);
    INTEnableSystemMultiVectoredInt();

	// get the peripheral bus speed
	fpb = FOSC / (1 << mOSCGetPBDIV());

	// Set up Timer 1 to generate interrupts at 1kHz
	// this is used to drive the underlying disk io system
    OpenCoreTimer(FOSC/ 2000);
    // set up the timer interrupt with a priority of 1
    mConfigIntCoreTimer(CT_INT_ON | CT_INT_PRIOR_1);

	// configure the pins of the SPI initially
	mPORTFClearBits(BIT_6 | BIT_8);
	mPORTFSetPinsDigitalOut(BIT_6 | BIT_8);
	_SetChipSelect(FALSE);
	
	// configure the input pins and change notification pullups
	mPORTDSetPinsDigitalIn(BIT_6 | BIT_7 | BIT_13);
	ConfigCNPullups(CN15_PULLUP_ENABLE | CN16_PULLUP_ENABLE | CN19_PULLUP_ENABLE);
	// configure the output LEDs
	mPORTDClearBits(BIT_0 | BIT_1 | BIT_2);
	mPORTDSetPinsDigitalOut(BIT_0 | BIT_1 | BIT_2);
	
	DBPRINTF("FatFs example application for PIC32\n");
	DBPRINTF("SW1 = Disk Status, SW2 = Dir, SW3 = Action Menu\n");
	
	// Main application 
	// a very simple command interface that allows us to perform a
	// number of operations on the SD card
	while (1) {
		// regulate the state machine operation
		while (tick < 100);  	// wait 100 ms
		tick = 0;
		
		switch (demoState) {
			case WAIT_FOR_DISK:
				if (!(disk_status(MMC_DRIVE) & STA_NODISK)) {
					// disk inserted so initialise it
					if (disk_initialize(MMC_DRIVE) == 0) {
						if (f_mount(MMC_DRIVE, &fatfs[MMC_DRIVE]) == FR_OK) {
							DBPRINTF("Disk initialized and mounted\n");
							demoState = IDLE;
						}	
					}
				}
				break;
				
			case IDLE:
				if (disk_status(MMC_DRIVE) & STA_NODISK) {
					demoState = WAIT_FOR_DISK;
					mPORTDClearBits(BIT_0);
					break;
				} else {
					// indicate card inserted using LED1 (RD0)
					mPORTDSetBits(BIT_0);
				}		
				
				// act upon button presses
				if (mPORTDReadBits(BIT_6) == 0) {
					// display status and directory information
					strcpy(buff, "1:");
					res = f_getfree(buff, &param1, &fs);
					if (res == FR_OK){
						DBPRINTF("\rFAT type = %u\rBytes/Cluster = %lu\rNumber of FATs = %u\r"
								"Root DIR entries = %u\rSectors/FAT = %lu\r"
								"Number of clusters = %lu\n",
								(WORD)fs->fs_type, (DWORD)fs->csize * 512, (WORD)fs->n_fats,
								fs->n_rootdir, fs->sects_fat, (DWORD)fs->max_clust - 2);
						DBPRINTF("\rFAT start (lba) = %lu\rDIR start (lba,clustor) = %lu\r"
								"Data start (lba) = %lu\n",
								fs->fatbase, fs->dirbase, fs->database);
					}
					
					acc_size = acc_files = acc_dirs = 0;
					res = scan_files(buff);
					if (res == FR_OK) { 
						DBPRINTF("\r%u files, %lu bytes.\r%u folders.\r"
								"%lu KB total disk space.\r%lu KB available.\n",
								acc_files, acc_size, acc_dirs,
								(fs->max_clust - 2) * (fs->csize / 2), 
								param1 * (fs->csize / 2));
					}					
					break;		
					DBPRINTF("SW1 = Disk Status, SW2 = Dir, SW3 = Action Menu\n");
				}
				
				// display a directory listing
				if (mPORTDReadBits(BIT_7) == 0) {
					strcpy(buff, "1:");
					res = f_opendir(&dir, buff);
					if (res == FR_OK) {
						param1 = s1 = s2 = 0;	
						
						while (1) {
							res = f_readdir(&dir, &finfo);	
							if ((res != FR_OK) || (!finfo.fname[0])) 
								break;
							if (finfo.fattrib & AM_DIR) {
								s2++;
							} else {
								s1++;
								param1 += finfo.fsize;
							}
							
							DBPRINTF("%c%c%c%c%c %u/%02u/%02u %02u:%02u %9lu  %s\n", 
									(finfo.fattrib & AM_DIR) ? 'D' : '-',
									(finfo.fattrib & AM_RDO) ? 'R' : '-',
									(finfo.fattrib & AM_HID) ? 'H' : '-',
									(finfo.fattrib & AM_SYS) ? 'S' : '-',
									(finfo.fattrib & AM_ARC) ? 'A' : '-',
									(finfo.fdate >> 9) + 1980, (finfo.fdate >> 5) & 15, finfo.fdate & 31,
									(finfo.ftime >> 11), (finfo.ftime >> 5) & 63,
									finfo.fsize, &(finfo.fname[0]));
						}					
						DBPRINTF("%4u File(s),%10lu bytes total\r%4u Dir(s)\n", s1, param1, s2);
						DBPRINTF("SW1 = Disk Status, SW2 = Dir, SW3 = Action Menu\n");
					}	
					break;		
				}					

				if (mPORTDReadBits(BIT_13) == 0) {
					// move on to the action menu
					demoState = ACTION;
					DBPRINTF("SW1 = Light LEDs, SW2 = Write Image, SW3 = Return\n");
					break;
				}

				break;
				
			case ACTION:
				// handle the action menu items
				if (mPORTDReadBits(BIT_6) == 0) {
					demoState = LED_PLAYBACK_OPEN;
					break;	
				}
				
				// write a bitmap image to the disk
				if (mPORTDReadBits(BIT_7) == 0) {
					res = f_open(&hFile, "1:PIC32.BMP", FA_CREATE_ALWAYS | FA_WRITE);
					if (res) {
						DBPRINTF("Failed to open file for writing\n");
						f_close(&hFile);
						break;
					}
	
					res = f_write(&hFile, bmpPIC32, sizeof(bmpPIC32), &cnt);
					if (res || (cnt != sizeof(bmpPIC32))) {
						DBPRINTF("Failed to write file\n");
					} else {
						DBPRINTF("File copied to disk\n");
					}
					f_close(&hFile);
					DBPRINTF("SW1 = Light LEDs, SW2 = Write Image, SW3 = Return\n");
					break;
				}
				
				// return to the main menu
				if (mPORTDReadBits(BIT_13) == 0) {
					demoState = IDLE;
					DBPRINTF("SW1 = Disk Status, SW2 = Dir, SW3 = Action Menu\n");
					break;
				}
												
				break;
			
			case LED_PLAYBACK_OPEN:
				DBPRINTF("Playing back led sequence in 'led.txt'\n");
				res = f_open(&hFile, "1:led.txt", FA_READ | FA_OPEN_EXISTING);
				if (res == 0) {
					demoState = LED_PLAYBACK_READ;
				} else {
					DBPRINTF("Failed to find file\n");
					demoState = IDLE;
					DBPRINTF("SW1 = Light LEDs, SW2 = Write Image, SW3 = Return\n");
				}
				break;
				
			case LED_PLAYBACK_READ:
				// read the next set of LED settings
				// a file created with DOS will have CR+LF pairs so
				// the length of each line will be 5 bytes
				res = f_read(&hFile, buff, 5, &cnt);
				if ((res != FR_OK) || (cnt != 5)) {
					// error or end of file
					demoState = LED_PLAYBACK_CLOSE;
					break;	
				}
				// output to the LEDs (forming the pattern)
				led_pattern = 0;
				if (buff[0] == '1') 
					led_pattern = 4;
				if (buff[1] == '1')
					led_pattern |= 2;
				if (buff[2] == '1')
					led_pattern |= 1;
				mPORTDClearBits(BIT_0 | BIT_1 | BIT_2);
				mPORTDSetBits(led_pattern);
				break;
			
			case LED_PLAYBACK_CLOSE:
				f_close(&hFile);
				demoState = ACTION;
				DBPRINTF("SW1 = Light LEDs, SW2 = Write Image, SW3 = Return\n");
				break;			
		}
	}
	
	return 0;	
}

/*********************************************************************
 * Function:  		void T1Handler(void)     
 *
 * PreCondition:    
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    
 *
 * Overview:        FatFs requires a 1ms tick timer to aid
 *					with low level function timinng
 *
 * Note:           
 ********************************************************************/
void __attribute__((interrupt(ipl1), vector(_CORE_TIMER_VECTOR))) T1Handler(void)
{
	static const BYTE dom[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	static int div1k;
	BYTE n;
	
	// clear the interrupt flag
	mCTClearIntFlag();
	OpenCoreTimer(FOSC/2000);
	Timer++;
	disk_timerproc();	// call the low level disk IO timer functions
	tick++;				// increment the local pace counter
	
	// implement a 'fake' RTCC
	if (++div1k >= 1000) {
		div1k = 0;
		if (++rtcSec >= 60) {
			rtcSec = 0;
			if (++rtcMin >= 60) {
				rtcMin = 0;
				if (++rtcHour >= 24) {
					rtcHour = 0;
					n = dom[rtcMon - 1];
					if ((n == 28) && !(rtcYear & 3)) n++;
					if (++rtcMday > n) {
						rtcMday = 1;
						if (++rtcMon > 12) {
							rtcMon = 1;
							rtcYear++;
						}
					}
				}
			}
		}		
	}
}


/*********************************************************************
 * Function:  		DWORD get_fattime(void)
 *
 * PreCondition:    
 *
 * Input:           None
 *
 * Output:          Time
 *
 * Side Effects:    
 *
 * Overview:        when writing fatfs requires a time stamp
 *					in this exmaple we are going to use a counter
 *					If the starter kit has the 32kHz crystal
 *					installed then the RTCC could be used instead
 *
 * Note:           
 ********************************************************************/
DWORD get_fattime(void)
{
	DWORD tmr;

	INTDisableInterrupts();
	tmr =	  (((DWORD)rtcYear - 80) << 25)
			| ((DWORD)rtcMon << 21)
			| ((DWORD)rtcMday << 16)
			| (WORD)(rtcHour << 11)
			| (WORD)(rtcMin << 5)
			| (WORD)(rtcSec >> 1);
	INTEnableInterrupts();

	return tmr;
}

///////////////////////////////////////////////////////////////////
// scan files in a given path for type information
static FRESULT scan_files(char* path)
{
	DIR dirs;
	FRESULT res;
	BYTE i;

	if ((res = f_opendir(&dirs, path)) == FR_OK) {
		i = strlen(path);
		while (((res = f_readdir(&dirs, &finfo)) == FR_OK) && finfo.fname[0]) {
			if (finfo.fattrib & AM_DIR) {
				acc_dirs++;
				*(path+i) = '/'; strcpy(path+i+1, &finfo.fname[0]);
				res = scan_files(path);
				*(path+i) = '\0';
				if (res != FR_OK) break;
			} else {
				acc_files++;
				acc_size += finfo.fsize;
			}
		}
	}
	return res;
}

